import e3nn
import pytest
from packaging.version import Version
import torch

from franken.config import BackboneConfig, GaussianRFConfig
from franken.data import BaseAtomsDataset
from franken.datasets.registry import DATASET_REGISTRY
from franken.backbones import REGISTRY
from franken.backbones.utils import load_checkpoint
from franken.rf.model import FrankenPotential


models = [
    "Egret-1t",
    pytest.param("MACE-L1", marks=pytest.mark.xfail(Version(e3nn.__version__) >= Version("0.5.5"), reason="Known incompatibility", strict=True)),
    pytest.param("MACE-OFF-small", marks=pytest.mark.xfail(Version(e3nn.__version__) >= Version("0.5.5"), reason="Known incompatibility", strict=True)),
    pytest.param("SevenNet0", marks=pytest.mark.xfail(Version(e3nn.__version__) < Version("0.5.0"), reason="Known incompatibility", strict=True)),
    pytest.param("SchNet-S2EF-OC20-200k", marks=pytest.mark.xfail(reason="Fails in CI due to unknown reasons", strict=False))
]


@pytest.mark.parametrize("model_name", models)
def test_backbone_loading(model_name):
    registry_entry = REGISTRY[model_name]
    gnn_config = BackboneConfig.from_ckpt({
        "family": registry_entry["kind"],
        "path_or_id": model_name,
        "interaction_block": 2,
    })
    load_checkpoint(gnn_config)


@pytest.mark.parametrize("model_name", models)
def test_descriptors(model_name):
    registry_entry = REGISTRY[model_name]
    gnn_config = BackboneConfig.from_ckpt({
        "family": registry_entry["kind"],
        "path_or_id": model_name,
        "interaction_block": 2,
    })
    bbone = load_checkpoint(gnn_config)
    # Get a random data sample
    data_path = DATASET_REGISTRY.get_path("test", "train", None, False)
    dataset = BaseAtomsDataset.from_path(
        data_path=data_path,
        split="train",
        gnn_config=gnn_config,
    )
    data, _ = dataset[0]  # type: ignore
    expected_fdim = bbone.feature_dim()
    features = bbone.descriptors(data)
    assert features.shape[1] == expected_fdim


@pytest.mark.parametrize("model_name", models)
def test_force_maps(model_name):
    from franken.backbones.wrappers.common_patches import patch_e3nn
    patch_e3nn()
    registry_entry = REGISTRY[model_name]
    gnn_config = BackboneConfig.from_ckpt({
        "family": registry_entry["kind"],
        "path_or_id": model_name,
        "interaction_block": 2,
    })
    # Get a random data sample
    data_path = DATASET_REGISTRY.get_path("test", "train", None, False)
    dataset = BaseAtomsDataset.from_path(
        data_path=data_path,
        split="train",
        gnn_config=gnn_config,
    )
    device="cuda:0" if torch.cuda.is_available() else "cpu"
    # initialize model
    model = FrankenPotential(
        gnn_config=gnn_config,
        rf_config=GaussianRFConfig(num_random_features=128, length_scale=1.0),
    )
    model = model.to(device)
    data, _ = dataset[0]  # type: ignore
    data = data.to(device)
    emap, fmap = model.grad_feature_map(data)




from typing import Final, Optional, Tuple
import mace
from packaging.version import Version

import torch
from mace.modules.models import MACE
from mace.modules.utils import get_edge_vectors_and_lengths
from e3nn.util.jit import compile_mode
from e3nn import o3

from franken.data import Configuration
from franken.utils.misc import torch_load_maybejit


def undo_script_mace(base: torch.jit.ScriptModule) -> MACE:
    from mace.modules.models import ScaleShiftMACE
    from mace.modules.blocks import (
        RealAgnosticResidualInteractionBlock,
        RealAgnosticInteractionBlock,
    )

    if base.original_name != "ScaleShiftMACE":
        raise ValueError("Only ScaleShiftMACE supported")

    def radial_to_name(radial_type):
        if radial_type == "BesselBasis":
            return "bessel"
        if radial_type == "GaussianBasis":
            return "gaussian"
        if radial_type == "ChebychevBasis":
            return "chebyshev"
        else:
            raise ValueError(f"Unrecognised bessel function {radial_type}")

    def radial_to_transform(radial):
        if not hasattr(radial, "distance_transform"):
            return None
        if radial.distance_transform.original_name == "AgnesiTransform":
            return "Agnesi"
        if radial.distance_transform.original_name == "SoftTransform":
            return "Soft"
        else:
            raise ValueError(
                f"Unrecognised distance transform {radial.distance_transform.original_name}"
            )

    def interaction_cls(module):
        if module.original_name == "RealAgnosticResidualInteractionBlock":
            return RealAgnosticResidualInteractionBlock
        elif module.original_name == "RealAgnosticInteractionBlock":
            return RealAgnosticInteractionBlock
        else:
            raise ValueError(f"Unrecognised interaction class {module.original_name}")

    def gate_fn(non_linearity):
        nonlin = non_linearity._modules["acts"]._modules["0"]
        if hasattr(nonlin, "f"):
            return nonlin.f
        code = nonlin.code
        if "silu" in code:
            return torch.nn.functional.silu
        elif "torch.abs" in code:
            return torch.abs
        else:
            raise ValueError(f"Unrecognized gating function. Code is {code}")

    module_dict = {nm[0]: nm[1] for nm in base.named_modules()}
    heads = base.heads if hasattr(base, "heads") else ["default"]
    num_interactions = int(base.num_interactions.item())
    model_mlp_irreps = (
        o3.Irreps(module_dict[f"readouts.{num_interactions - 1}"].hidden_irreps)
        if base.num_interactions.item() > 1
        else 1
    )
    try:
        correlation = (
            len(
                list(
                    module_dict[
                        "products.0.symmetric_contractions.contractions.0"
                    ].weights.parameters()
                )
            )
            + 1
        )
    except (KeyError, AttributeError):
        correlation = module_dict[
            "products.0.symmetric_contractions"
        ].contraction_degree

    args = {}
    args["r_max"] = base.r_max.item()
    args["num_bessel"] = len(base.radial_embedding.bessel_fn.bessel_weights)
    args["num_polynomial_cutoff"] = base.radial_embedding.cutoff_fn.p.item()
    args["max_ell"] = base.spherical_harmonics._lmax
    args["interaction_cls"] = interaction_cls(module_dict["interactions.1"])
    args["interaction_cls_first"] = interaction_cls(module_dict["interactions.0"])
    args["num_interactions"] = num_interactions
    args["num_elements"] = len(base.atomic_numbers)
    args["hidden_irreps"] = o3.Irreps(module_dict["products.0.linear"].irreps_out)
    # args["edge_irreps"] = base.edge_irreps if hasattr(base, "edge_irreps") else None
    args["MLP_irreps"] = (
        o3.Irreps(f"{model_mlp_irreps.count((0, 1)) // len(heads)}x0e")  # type: ignore
        if num_interactions > 1
        else 1
    )
    args["atomic_energies"] = base.atomic_energies_fn.atomic_energies.cpu().numpy()
    args["avg_num_neighbors"] = module_dict["interactions.0"].avg_num_neighbors
    args["atomic_numbers"] = base.atomic_numbers.cpu().numpy()
    args["correlation"] = correlation
    args["gate"] = (
        gate_fn(module_dict[f"readouts.{num_interactions - 1}"].non_linearity)
        if num_interactions > 1
        else None
    )
    args["pair_repulsion"] = hasattr(base, "pair_repulsion")
    args["distance_transform"] = radial_to_transform(base.radial_embedding)
    args["radial_MLP"] = module_dict["interactions.0"].conv_tp_weights.hs[1:-1]
    # args["radial_MLP"] = module_dict["interactions.0"].radial_MLP
    args["radial_type"] = radial_to_name(base.radial_embedding.bessel_fn.original_name)
    args["heads"] = base.heads
    # args["use_reduced_cg"] = base.use_reduced_cg if hasattr(base, "use_reduced_cg") else False
    # args["use_so3"] = base.use_so3 if hasattr(base, "use_so3") else False
    # args["cueq_config"] = base.cueq_config if hasattr(base, "cueq_config") else None
    args["atomic_inter_scale"] = base.scale_shift.scale.cpu().numpy()
    args["atomic_inter_shift"] = base.scale_shift.shift.cpu().numpy()
    model = ScaleShiftMACE(**args)
    load_result = model.load_state_dict(base.state_dict(), strict=False)
    if len(load_result.unexpected_keys) > 0:
        raise RuntimeError(
            f"Failed to load state-dict. There were unexpected keys: {load_result.unexpected_keys}"
        )
    for k in load_result.missing_keys:
        if (
            "weights_0_zeroed" not in k
            and "weights_1_zeroed" not in k
            and "weights_max_zeroed" not in k
        ):
            raise RuntimeError(
                f"Failed to load state-dict. There were missing keys: {load_result.missing_keys}"
            )
    return model


@compile_mode("script")
class FrankenMACE(torch.nn.Module):
    """Wrap MACE module.

    Most of the complication here is to support torch-scripted modules, which don't play well
    with `torch.func.jacfwd` and other transformations. In particular
     - the spherical harmonics of e3nn don't work as they are jitted by default. We have a patch
        in :module:`franken.backbones.wrappers.common_patches` which fixes this, but does not get
        called when the full model is scripted.
     - the radial embedding doesn't work (more investigation needed to figure out exactly which
        part of it fails), so we have to rebuild it discarding the jitted version.
     - In order to extract features at a specific layer, we also have to pull apart the interactions
        and products module-lists which are not easily indexable when jitted.
    """

    interaction_block: Final[int]

    def __init__(self, base_model: MACE, interaction_block, gnn_backbone_id):
        super().__init__()
        self.gnn_backbone_id = gnn_backbone_id
        self.interaction_block = interaction_block
        # Copy things from base model
        if isinstance(base_model, torch.jit.ScriptModule):
            base_model = undo_script_mace(base_model)
        self.atomic_numbers = base_model.atomic_numbers
        self.r_max = base_model.r_max
        self.register_buffer(
            "num_interactions", torch.tensor(interaction_block, dtype=torch.int64)
        )
        self.node_embedding = base_model.node_embedding
        self.spherical_harmonics = base_model.spherical_harmonics
        self.radial_embedding = base_model.radial_embedding
        self.interactions = base_model.interactions[: self.interaction_block]
        self.products = base_model.products[: self.interaction_block]
        self.is_mace_v3_14 = Version(mace.__version__) >= Version("0.3.14")  # type: ignore
        self.is_mace_v3_13 = Version(mace.__version__) >= Version("0.3.13")  # type: ignore

    def init_args(self):
        return {
            "gnn_backbone_id": self.gnn_backbone_id,
            "interaction_block": self.interaction_block,
        }

    def descriptors(self, data: Configuration) -> torch.Tensor:
        # assert on local variables to make torchscript happy
        edge_index = data.edge_index
        shifts = data.shifts
        node_attrs = data.node_attrs
        assert edge_index is not None
        assert shifts is not None
        assert node_attrs is not None
        # Embeddings
        node_feats = self.node_embedding(node_attrs)  # type: ignore
        vectors, lengths = get_edge_vectors_and_lengths(
            positions=data.atom_pos,
            edge_index=edge_index,
            shifts=shifts,
        )
        edge_attrs = self.spherical_harmonics(vectors)  # type: ignore
        rad_emb = self.radial_embedding(
            lengths, node_attrs, edge_index, self.atomic_numbers
        )  # type: ignore
        if torch.jit.isinstance(rad_emb, Tuple[torch.Tensor, Optional[float]]):
            edge_feats, cutoff = rad_emb
        else:
            edge_feats = rad_emb
            cutoff = None

        # lammps_class, lammps_natoms are only set when using LAMMPS MLIAP
        # the defaults are used otherwise and set here
        lammps_class = None
        lammps_natoms = (0, 0)

        node_feats_list = []
        for i, (interaction, product) in enumerate(
            zip(self.interactions, self.products)
        ):
            if self.is_mace_v3_14:
                node_feats, sc = interaction(
                    node_attrs=node_attrs,
                    node_feats=node_feats,
                    edge_attrs=edge_attrs,
                    edge_feats=edge_feats,
                    edge_index=edge_index,
                    cutoff=cutoff,
                    first_layer=(i == 0),
                    lammps_class=lammps_class,
                    lammps_natoms=lammps_natoms,
                )  # type: ignore
            elif self.is_mace_v3_13:
                node_feats, sc = interaction(
                    node_attrs=node_attrs,
                    node_feats=node_feats,
                    edge_attrs=edge_attrs,
                    edge_feats=edge_feats,
                    edge_index=edge_index,
                    first_layer=(i == 0),
                    lammps_class=lammps_class,
                    lammps_natoms=lammps_natoms,
                )  # type: ignore
            else:
                node_feats, sc = interaction(
                    node_attrs=node_attrs,
                    node_feats=node_feats,
                    edge_attrs=edge_attrs,
                    edge_feats=edge_feats,
                    edge_index=edge_index,
                )  # type: ignore

            node_feats = product(
                node_feats=node_feats, sc=sc, node_attrs=node_attrs
            )  # type: ignore
            # Extract only scalars. Use `irreps_out` attribute to figure out which features correspond to scalars.
            # irreps_out is an `Irreps` object: a 2-tuple of multiplier and `Irrep` objects.
            # Tuple[Tuple[int, Tuple[int, int]], Tuple[int, Tuple[int, int]]]
            # The `Irrep` object is a tuple consisting of parameters `l` and `p`.
            # The scalar irrep is the first in `irreps_out`. Its dimension is computed
            # as `mul * ir.dim` where `ir.dim == 2 * ir.l  + 1`
            # Note this is equivalent code, which does not support TorchScript.
            # invariant_slices = product.linear.irreps_out.slices()[0]
            irreps = product.linear.irreps_out
            invariant_slices = slice(0, irreps[0][0] * (2 * irreps[0][1][0] + 1))
            node_feats_list.append(node_feats[..., invariant_slices])
        return torch.cat(node_feats_list, dim=-1)

    def num_params(self) -> int:
        return sum(p.numel() for p in self.parameters())

    def feature_dim(self):
        nfeat = 0
        for p in self.products:
            nfeat += p.linear.irreps_out[0][0] * (2 * p.linear.irreps_out[0][1][0] + 1)
        return nfeat

    @staticmethod
    def load_from_checkpoint(
        trainer_ckpt, gnn_backbone_id: str, interaction_block: int, map_location=None
    ) -> "FrankenMACE":
        mace = torch_load_maybejit(
            trainer_ckpt, map_location=map_location, weights_only=False
        ).to(dtype=torch.float32)
        return FrankenMACE(
            base_model=mace,
            gnn_backbone_id=gnn_backbone_id,
            interaction_block=interaction_block,
        )



import logging

import torch


logger = logging.getLogger("franken")


def patch_e3nn():
    # NOTE:
    #  Patching should occur during training: it is necessary for `jvp` on the MACE model,
    #  but not during inference, when we only use `torch.autograd`. For inference, we may want
    #  to compile the model using `torch.jit` - and the patch interferes with the JIT, so we
    #  must disable it.

    import e3nn.o3._spherical_harmonics

    if hasattr(e3nn.o3._spherical_harmonics._spherical_harmonics, "code"):
        # Then _spherical_harmonics is a scripted function, we need to undo this!
        new_locals = {"Tensor": torch.Tensor}
        exec(e3nn.o3._spherical_harmonics._spherical_harmonics.code, None, new_locals)

        def _spherical_harmonics(
            lmax: int, x: torch.Tensor, y: torch.Tensor, z: torch.Tensor
        ) -> torch.Tensor:
            return new_locals["_spherical_harmonics"](torch.tensor(lmax), x, y, z)

        # Save to allow undoing later
        setattr(
            e3nn.o3._spherical_harmonics,
            "_old_spherical_harmonics",
            e3nn.o3._spherical_harmonics._spherical_harmonics,
        )
        e3nn.o3._spherical_harmonics._spherical_harmonics = _spherical_harmonics

    # 2nd patch for newer e3nn versions (somewhere between 0.5.0 and 0.5.5
    # e3nn jits _spherical_harmonics which the SphericalHarmonics class,
    # making the above patch ineffective)
    try:
        from e3nn import set_optimization_defaults

        set_optimization_defaults(jit_script_fx=False)
    except ImportError:
        pass  # only valid for newer e3nn


def unpatch_e3nn():
    # This is only useful for CI and testing environments.
    # When jit-compiling a franken module (e.g. for LAMMPS), we don't want the patch applied!
    import e3nn.o3._spherical_harmonics

    if hasattr(e3nn.o3._spherical_harmonics, "_old_spherical_harmonics"):
        e3nn.o3._spherical_harmonics._spherical_harmonics = (
            e3nn.o3._spherical_harmonics._old_spherical_harmonics
        )



import hashlib
import logging
import math
from pathlib import Path
from time import perf_counter
from typing import Any, List, Literal, Mapping, Sequence

import numpy as np
import torch
import torch.utils.data
from torch import Tensor

import franken.metrics
from franken.metrics.base import BaseMetric
import franken.utils.distributed as dist_utils
from franken.data.base import Target
from franken.rf.model import FrankenPotential
from franken.trainers import BaseTrainer
from franken.trainers.log_utils import (
    DataSplit,
    HyperParameterGroup,
    LogCollection,
    LogEntry,
)
from franken.utils.linalg.cov import (
    lowmem_normalize_leading_eig,
    rank1_update,
    rankk_update,
)
from franken.utils.linalg.psdsolve import psd_ridge
from franken.utils.linalg.tri import triangular_lerp
from franken.utils.misc import params_grid, throughput


logger = logging.getLogger("franken")


class RandomFeaturesTrainer(BaseTrainer):
    """Main class which groups training and evaluation functionality for franken models.

    Args:
        train_dataloader (torch.utils.data.DataLoader):
            Dataloader which iterates over the training set.
        random_features_normalization (Literal["leading_eig"] | None):
            How to normalize the covariance matrices formed by random-features. Defaults to "leading_eig".
        log_dir (Path | None):
            Directory where to save logs and models. If not specified, no logs will be saved.
            Defaults to None.
        save_every_model (bool):
            Model fitting with this class is done simultaneously for a list
            of solver parameters. This argument controls the behavior of model saving:
            if set to True, the models corresponding to all solver parameters will be saved,
            otherwise only the 'best' model among them (according to some validation set) will
            be saved. Defaults to True.
        device:
            PyTorch device on which computations are performed. Defaults to "cuda:0".
        dtype (str | torch.dtype):
            Data-type for solver operations. Random features will be computed in float32, and
            then converted to float64 if requested. Defaults to torch.float32.
        save_fmaps (bool):
            Whether or not to save feature-maps for the training set. Saving them
            requires extra memory (linear in the training-set size), but speeds up
            the :meth:`~franken.trainers.FrankenPotential.evaluate` method on training
            data. Defaults to True.
    """

    def __init__(
        self,
        train_dataloader: torch.utils.data.DataLoader,
        random_features_normalization: Literal["leading_eig"] | None = "leading_eig",
        log_dir: Path | None = None,
        save_every_model: bool = True,
        device: torch.device | str | int = "cuda:0",
        dtype: str | torch.dtype = torch.float32,
        save_fmaps: bool = True,
    ):
        super().__init__(
            train_dataloader,
            log_dir=log_dir,
            save_every_model=save_every_model,
            device=device,
            dtype=dtype,
        )
        self.random_features_normalization = random_features_normalization
        self.save_fmaps = save_fmaps

    def on_fit_start(self, model: FrankenPotential):
        # initialize input scaler based on statistics property
        model.input_scaler.set_from_statistics(self.get_statistics(model)[0])
        # initialize energy shift based on atomic energies
        if not model.energy_shift.is_initialized:
            model.energy_shift.set_from_atomic_energies(
                self.train_dataloader.dataset.atomic_energies
            )

    def fit(
        self,
        model: FrankenPotential,
        solver_params: Mapping[str, Sequence[Any]],
    ) -> tuple[LogCollection, torch.Tensor]:
        """Fit a given franken model on the training set.

        Args:
            model (FrankenPotential): The model which defines GNN and random features.
            solver_params (dict): Parameters for the solver which actually
                performs the fit. This argument allows to specify multiple parameters,
                for each of which we will perform a fit. For example

                >>> solver_params = {
                >>>     "l2_penalty": [1e-6, 1e-4],
                >>>     "force_weight": [0.5]
                >>> }

                will result in two different models, one with :code:`l2_penalty=1e-6, force_weight=0.5`
                and one with :code:`l2_penalty=1e-4, force_weight=0.5`. This way of specifying solver
                parameters allows to easily perform a grid-search.

        Returns:
            logs (LogCollection): Logs which contain all parameters related
                to the fitting, as well as timings.
            weights (torch.Tensor): Weights which were learned during the fit.

        Note:
            More information about the available solver parameters can be found under the
            :meth:`solve` method.
        """
        if self.device.type == "cuda":
            # Patch E3NN for batched jacobians!
            from franken.backbones.wrappers.common_patches import patch_e3nn

            patch_e3nn()

        model = model.to(self.device)
        self.on_fit_start(model)
        model_hash = hashlib.md5(str(model.hyperparameters).encode())
        model_hash = model_hash.hexdigest()

        t_cov_coeffs_start = perf_counter()
        self._compute_covs_and_coeffs(model, self.train_dataloader)
        t_cov_coeffs = perf_counter() - t_cov_coeffs_start

        solver_grid_size = math.prod([len(v) for v in solver_params.values()])
        all_weights = torch.zeros(
            solver_grid_size,
            model.rf.total_random_features,
            dtype=self.buffer_dt,
            device=self.device,
        )

        solver_iter = throughput(
            params_grid(solver_params, split_distributed=True),
            desc="least-squares",
            units="models",
            device=self.device,
        )

        local_logs = dict()
        num_failed = torch.zeros((1,), device=self.device, dtype=torch.int)
        for hp_idx, hp_val in solver_iter:
            t_solve_start = perf_counter()
            try:
                weights = self.solve(**hp_val)
            except torch.linalg.LinAlgError as e:
                weights = torch.full_like(all_weights[hp_idx], torch.inf)
                num_failed += 1
                logger.debug(f"Hyperparameter {hp_val} failed. Error: {e}")
            finally:
                t_solve = perf_counter() - t_solve_start
            # Update all weights
            all_weights[hp_idx].copy_(weights.view(-1))

            # Logging
            solver_hps = hp_val | {"dtype": self.buffer_dt}
            hp_groups = model.hyperparameters | {"solver": solver_hps}
            hyperparameters = []
            for group_name, hps in hp_groups.items():
                hyperparameters.append(HyperParameterGroup.from_dict(group_name, hps))

            local_logs[hp_idx] = LogEntry(
                checkpoint_hash=model_hash,
                checkpoint_rf_weight_id=hp_idx,
                timings_cov_coeffs=t_cov_coeffs,
                timings_solve=t_solve,
                hyperparameters=hyperparameters,
            )

        # Broadcast weights and logs
        dist_utils.all_sum(all_weights)
        log_collection = LogCollection.gather_from_ranks(local_logs)
        assert len(log_collection) == solver_grid_size

        # Reporting failed runs
        dist_utils.all_sum(num_failed)

        if num_failed.item() > 0:
            logger.warning(
                f"Solver failed in {num_failed.item()}/{solver_grid_size} cases."
            )

        return log_collection, all_weights

    def evaluate(
        self,
        model: FrankenPotential,
        dataloader: torch.utils.data.DataLoader,
        log_collection: LogCollection,
        all_weights: torch.Tensor | None,
        metrics: Sequence[str] = ("energy_MAE", "forces_MAE", "forces_cosim"),
    ) -> LogCollection:
        if self.device.type == "cuda":
            # Patch E3NN for batched jacobians!
            from franken.backbones.wrappers.common_patches import patch_e3nn

            patch_e3nn()
        tot_dset_size = len(dataloader.dataset)  # type: ignore

        metric_objects: List[BaseMetric] = []
        for name in metrics:
            try:
                metric_obj = franken.metrics.init_metric(
                    name, device=self.device, dtype=self.buffer_dt
                )
                metric_objects.append(metric_obj)
            except KeyError:
                logger.warning(
                    f"Unknown metric {name}. Skipping. Available metrics: {franken.metrics.available_metrics()}"
                )

        split_name = dataloader.dataset.split
        try:
            split = DataSplit[split_name.upper()]
        except KeyError:
            logger.warning(f"Unrecognized split '{split_name}' in dataloader")
            split = DataSplit.UNDEFINED

        progress_bar = throughput(
            dataloader,
            desc=f"{split.name.lower()} evaluation",
            total=tot_dset_size,
            device=self.device,
        )
        for i, (data, targets) in enumerate(progress_bar):
            data = data.to(device=self.device)
            targets = targets.to(device=self.device)
            if split == DataSplit.TRAIN and self.save_fmaps:
                # Shortcut to compute predictions for the training-set, for which
                # we already have computed energy and force feature maps
                assert len(self.forces_fmap) == len(self.energy_fmap)
                assert len(self.forces_fmap) == len(dataloader)
                predictions = Target(
                    *model.energy_and_forces_from_fmaps(
                        data,
                        energy_fmap=self.energy_fmap[i],
                        forces_fmap=self.forces_fmap[i],
                        weights=all_weights,
                        add_energy_shift=False,  # since it's always train here
                    )
                )
            else:
                if all_weights is None or all_weights.shape[0] <= 100:
                    forces_mode = "torch.autograd"
                else:
                    forces_mode = "torch.func"
                predictions = Target(
                    *model.energy_and_forces(
                        data,
                        weights=all_weights,
                        forces_mode=forces_mode,
                        add_energy_shift=(False if split == DataSplit.TRAIN else True),
                    )
                )
            if torch.any(torch.isnan(predictions.energy)):
                logger.warning(
                    f"Configuration {i} - {split_name} has NaNs in energy predictions"
                )
            if predictions.forces is not None and torch.any(
                torch.isnan(predictions.forces)
            ):
                logger.warning(
                    f"Configuration {i} - {split_name} has NaNs in force predictions"
                )
            for metric in metric_objects:
                metric.update(predictions, targets)

        num_models = (
            all_weights.shape[0]
            if all_weights is not None
            else model.rf.weights.shape[0]
        )
        # Sync metrics across GPUs
        metric_values = {}
        for metric in metric_objects:
            value = metric.compute(reset=False)
            if metric.samples_counter.item() != tot_dset_size:
                logger.warning(
                    f"Metric {metric.name} has {metric.samples_counter.item()} samples "
                    f"while the dataset has {tot_dset_size} configurations."
                )
            assert value.ndim == 1
            assert value.shape[0] == num_models
            metric_values[metric.name] = value

        # Explode metric values into list of MetricLog
        raw_logs = []
        for idx in range(num_models):
            results = []
            for name, value in metric_values.items():
                float_value = value[idx].item()
                results.append(dict(name=name, value=float_value))
            raw_logs.append(results)

        assert len(raw_logs) == len(log_collection)

        for log_entry, results in zip(log_collection, raw_logs):
            for metric in results:
                try:
                    log_entry.add_metric(metric["name"], metric["value"], split)
                except ValueError as e:
                    logger.warning(f"Could not add metric because: {str(e)}")
        return log_collection

    def warn_save_fmaps(
        self, energy_fmap: Tensor, forces_fmap: Tensor, num_maps: int
    ) -> None:
        fmap_size = np.prod(energy_fmap.shape) + np.prod(forces_fmap.shape)
        tot_bytes = fmap_size * forces_fmap.element_size() * num_maps
        if self.device.type == "cuda":
            avail_bytes = torch.cuda.mem_get_info(self.device)[0]
            if tot_bytes > 0.8 * avail_bytes:
                logger.warning(
                    f"Saved feature maps require {tot_bytes / 2**30:.2f}GB of device memory. "
                    f"Device has {avail_bytes / 2**30:.2f}GB of available memory, a crash may occur. "
                    f"Use 'trainer.save_fmaps=False' in your config to stop saving feature maps."
                )
            else:
                logger.info(
                    f"Saved feature maps require {tot_bytes / 2**30:.2f}GB of device memory."
                )

    @torch.no_grad()
    def _compute_covs_and_coeffs(
        self, model: FrankenPotential, dataloader: torch.utils.data.DataLoader
    ) -> None:
        tot_dset_size = len(dataloader.dataset)  # type: ignore
        n_rf = model.rf.total_random_features
        # Initialize Buffers: `covariance` will have
        # - covariance of forces on LOWER triangle
        # - covariance of energies on UPPER triangle
        # with diagonals stored separately, and linsys coefficients.
        self.covariance = torch.zeros(
            (n_rf, n_rf), device=self.device, dtype=self.buffer_dt
        )
        self.diag_energy = torch.zeros(
            (n_rf,), device=self.device, dtype=self.buffer_dt
        )
        self.diag_forces = torch.zeros(
            (n_rf,), device=self.device, dtype=self.buffer_dt
        )
        self.coeffs_energy = torch.zeros(
            (n_rf,), device=self.device, dtype=self.buffer_dt
        )
        self.coeffs_forces = torch.zeros(
            (n_rf,), device=self.device, dtype=self.buffer_dt
        )
        # NOTE: the feature maps are never synced between devices.
        #       They can only used correctly by iterating through the
        #       same dataloader as here, using the same method. Otherwise
        #       they may not be in the correct order
        self.energy_fmap = []
        self.forces_fmap = []

        progress_bar = throughput(
            dataloader,
            desc="covs+coeffs",
            total=tot_dset_size,
            device=self.device,
        )

        for i, (data, targets) in enumerate(progress_bar):
            data = data.to(device=self.device)
            targets = targets.to(device=self.device)

            energy_per_atom = targets.energy / data.natoms
            forces_per_atom = targets.forces / data.natoms

            forces_fmap, energy_fmap = model.grad_feature_map(data)
            energy_fmap = energy_fmap.to(dtype=self.buffer_dt)

            rank1_update(self.covariance, self.diag_energy, energy_fmap, upper=True)
            self.coeffs_energy.add_(energy_fmap, alpha=energy_per_atom.item())

            forces_fmap = (-forces_fmap).to(dtype=self.buffer_dt)
            forces_fmap = forces_fmap.view(forces_fmap.shape[0], -1)
            rankk_update(self.covariance, self.diag_forces, forces_fmap, upper=False)
            self.coeffs_forces.addmv_(
                forces_fmap, forces_per_atom.view(-1).to(dtype=self.buffer_dt)
            )
            if self.save_fmaps:
                if i == 0:
                    self.warn_save_fmaps(energy_fmap, forces_fmap, len(dataloader))
                self.energy_fmap.append(energy_fmap)
                self.forces_fmap.append(forces_fmap)

        # Sync covariance matrices & coefficients
        dist_utils.all_sum(self.covariance)
        dist_utils.all_sum(self.diag_forces)
        dist_utils.all_sum(self.diag_energy)
        dist_utils.all_sum(self.coeffs_energy)
        dist_utils.all_sum(self.coeffs_forces)

        if self.random_features_normalization == "leading_eig":
            logger.warning(
                "`leading_eig` normalization has high memory usage. If you encounter OOM errors try to disable it."
            )
            lowmem_normalize_leading_eig(
                self.covariance, self.diag_energy, self.coeffs_energy, upper=True
            )
            lowmem_normalize_leading_eig(
                self.covariance, self.diag_forces, self.coeffs_forces, upper=False
            )
        elif self.random_features_normalization is not None:
            raise ValueError(
                f"Covariance normalization {self.random_features_normalization} is not implemented."
            )

    @torch.no_grad()
    def solve(self, force_weight: float, l2_penalty: float = 1e-6) -> Tensor:
        # This is the 2nd copy of the covariance matrix that we need to store.
        lerped_cov, lerped_diag = triangular_lerp(
            self.covariance,
            diag_upper=self.diag_energy,
            diag_lower=self.diag_forces,
            weight=force_weight,
            inplace=False,
        )
        lerped_cov.diagonal().copy_(lerped_diag)
        rhs = torch.lerp(self.coeffs_energy, self.coeffs_forces, force_weight)
        solution = psd_ridge(lerped_cov, rhs, l2_penalty)
        return solution



import hashlib
from pathlib import Path
import requests
from tqdm.auto import tqdm


def compute_md5(file_path):
    md5_hash = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(
            lambda: f.read(8192), b""
        ):  # Read in chunks to handle large files
            md5_hash.update(chunk)
    return md5_hash.hexdigest()


def validate_file(
    filename: Path, expected_size: int | None = None, expected_md5: str | None = None
) -> bool:
    if expected_md5 is not None:
        actual_md5 = compute_md5(filename)
        if actual_md5 != expected_md5:
            return False
    if expected_size is not None:
        actual_size = filename.stat().st_size
        if expected_size != actual_size:
            return False
    return True


def download_file(
    url: str,
    filename: Path,
    expected_size: int | None = None,
    expected_md5: str | None = None,
    desc: str | None = None,
):
    if (expected_md5 is not None or expected_size is not None) and filename.is_file():
        # Check that the file is correct to avoid re-download
        if validate_file(filename, expected_size, expected_md5):
            return filename

    response = requests.get(url, stream=True)
    response.raise_for_status()
    total_size = int(response.headers.get("content-length", 0))
    block_size = 8192
    data_size = 0
    data_md5 = hashlib.md5()
    with (
        open(filename.with_suffix(".temp"), "wb") as file,
        tqdm(
            desc=desc or str(filename),
            total=total_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as bar,
    ):
        for chunk in response.iter_content(chunk_size=block_size):
            file.write(chunk)
            data_md5.update(chunk)
            data_size += len(chunk)
            bar.update(len(chunk))
    # validate
    if expected_size is not None and expected_size != data_size:
        raise IOError("Incorrect file size", filename)
    if expected_md5 is not None and data_md5.hexdigest() != expected_md5:
        raise IOError("Incorrect file MD5", filename)

    filename.with_suffix(".temp").replace(filename)
    return filename



from importlib import resources
from pathlib import Path

import franken.datasets
from franken.datasets.registry import DATASET_REGISTRY, BaseRegisteredDataset


@DATASET_REGISTRY.register("test")
class TestRegisteredDataset(BaseRegisteredDataset):
    relative_paths = {
        "test": {
            "train": "test/train.xyz",
            "val": "test/validation.xyz",
            "test": "test/test.xyz",
            "md": "test/md.xyz",
            "long": "test/long.xyz",
        },
    }

    @classmethod
    def get_path(
        cls, name: str, split: str, base_path: Path | None, download: bool = True
    ):
        relative_path = cls.relative_paths[name][split]
        path = resources.files(franken.datasets) / relative_path
        if path.is_file():
            return path
        else:
            raise ValueError(f"Dataset not found at '{path}'")
